class new_feature:
    def __init__(self, id, featuretype, seqid, source, strand, start, end, frame, attributes):
        self.id = id
        self.featuretype = featuretype
        self.seqid = seqid
        self.source = source
        self.strand = strand
        self.start = start
        self.end = end
        self.frame = frame
        self.attributes= attributes
